@@ -11,15 +11,14 @@ Provides the definition of the :class:`~ModelLink` abstract base class.
 # %% IMPORTS
 # Built-in imports
 import abc
+from glob import glob
 import os
 from os import path
 from tempfile import mktemp
 import warnings
 
 # Package imports
-from e13tools import InputError, ShapeError
-from e13tools.utils import (convert_str_seq, docstring_substitute,
-                            get_outer_frame, raise_error)
+import e13tools as e13
 import h5py
 import hickle
 import numpy as np
@@ -28,8 +27,8 @@ from sortedcontainers import SortedDict as sdict, SortedSet as sset
 # PRISM imports
 from prism import __version__
 from prism._docstrings import std_emul_i_doc
-from prism._internal import (FeatureWarning, RequestWarning, check_vals,
-                             getCLogger, np_array)
+from prism._internal import (
+    FeatureWarning, RequestWarning, check_vals, getCLogger, np_array)
 from prism.modellink.utils import convert_data, convert_parameters
 
 # All declaration
@@ -160,9 +159,9 @@ class ModelLink(object, metaclass=abc.ABCMeta):
         (linear, log, ln) the data values are. It defaults to 'lin' if it is
         not provided.
 
-        The data identifier is a sequence of ints, floats and strings that is
-        unique for every data point. *PRISM* uses it to identify a data point
-        with, which is required in some cases (like MPI), while the model
+        The data identifier is a sequence of bools, ints, floats and strings
+        that is unique for every data point. *PRISM* uses it to identify a data
+        point with, which is required in some cases (like MPI), while the model
         itself can use it as a description of the operations required to
         extract the data point from the model output. It can be provided as any
         sequence of any length for any data point. If any sequence contains a
@@ -207,39 +206,10 @@ class ModelLink(object, metaclass=abc.ABCMeta):
     # Define the representation of a ModelLink object
     def __repr__(self):
         # Obtain representation of model_parameters
-        par_repr = []
-        for name, rng, est in zip(self._par_name, self._par_rng,
-                                  self._par_est):
-            if est is None:
-                par_repr.append("%r: %r" % (name, rng))
-            else:
-                par_repr.append("%r: %r" % (name, [*rng, est]))
-        par_repr = "model_parameters={%s}" % (", ".join(map(str, par_repr)))
+        par_repr = "model_parameters=%s" % (dict(self.model_parameters))
 
         # Obtain representation of model_data
-        data_repr = []
-        data_points = []
-
-        # Combine data points together, only adding non-default values
-        for val, err, spc in zip(self._data_val, self._data_err,
-                                 self._data_spc):
-            # Add data value
-            data_points.append([val])
-
-            # Add data error, add only one value if error is centered
-            if(err[0] == err[1]):
-                data_points[-1].append(err[0])
-            else:
-                data_points[-1].extend(err)
-
-            # Add data space if it is not 'lin'
-            if(spc != 'lin'):
-                data_points[-1].append(spc)
-
-        # Combine data points with data identifiers
-        for idx, point in zip(self._data_idx, data_points):
-            data_repr.append("%r: %r" % (idx, point))
-        data_repr = "model_data={%s}" % (", ".join(map(str, data_repr)))
+        data_repr = "model_data=%s" % (self.model_data)
 
         # Obtain non-default representation and add default ones
         str_repr = self.get_str_repr()
@@ -265,7 +235,7 @@ class ModelLink(object, metaclass=abc.ABCMeta):
     @name.setter
     def name(self, name):
         # If name is set outside of __init__, save current value
-        outer_frame = get_outer_frame(self.__init__)
+        outer_frame = e13.get_outer_frame(self.__init__)
         if outer_frame is None and not hasattr(self, '_init_name'):
             self._init_name = str(self._name)
 
@@ -320,7 +290,7 @@ class ModelLink(object, metaclass=abc.ABCMeta):
     @call_type.setter
     def call_type(self, call_type):
         # If call_type is set outside of __init__, save current value
-        outer_frame = get_outer_frame(self.__init__)
+        outer_frame = e13.get_outer_frame(self.__init__)
         if outer_frame is None and not hasattr(self, '_init_call_type'):
             self._init_call_type = str(self._call_type)
 
@@ -367,7 +337,7 @@ class ModelLink(object, metaclass=abc.ABCMeta):
     @MPI_call.setter
     def MPI_call(self, MPI_call):
         # If MPI_call is set outside of __init__, save current value
-        outer_frame = get_outer_frame(self.__init__)
+        outer_frame = e13.get_outer_frame(self.__init__)
         if outer_frame is None and not hasattr(self, '_init_MPI_call'):
             self._init_MPI_call = bool(self._MPI_call)
 
@@ -405,12 +375,12 @@ class ModelLink(object, metaclass=abc.ABCMeta):
     @property
     def par_rng(self):
         """
-        :obj:`~numpy.ndarray`: The lower and upper values of the model
+        dict of :obj:`~numpy.ndarray`: The lower and upper values of the model
         parameters.
 
         """
 
-        return(self._par_rng)
+        return(sdict(zip(self._par_name, self._par_rng)))
 
     @property
     def par_est(self):
@@ -423,6 +393,33 @@ class ModelLink(object, metaclass=abc.ABCMeta):
 
         return(sdict(zip(self._par_name, self._par_est)))
 
+    @property
+    def model_parameters(self):
+        """
+        dict of list: The dict of model parameters as used by this
+        :obj:`~ModelLink` instance.
+
+        This dict can be used as the `model_parameters` argument when
+        initializing this :class:`~ModelLink` subclass.
+
+        """
+
+        # Initialize empty dict of model parameters
+        model_parameters = sdict()
+
+        # Loop over all parameter properties and add them to the dict
+        for name, rng, est in zip(self._par_name, self._par_rng,
+                                  self._par_est):
+            # If estimate was not given, only add the parameter range
+            if est is None:
+                model_parameters[name] = [*rng]
+            # Else, add range and estimate
+            else:
+                model_parameters[name] = [*rng, est]
+
+        # Return model_parameters
+        return(model_parameters)
+
     # Model Data
     @property
     def n_data(self):
@@ -436,31 +433,31 @@ class ModelLink(object, metaclass=abc.ABCMeta):
     @property
     def data_val(self):
         """
-        list of float: The values of provided data points.
+        dict of float: The values of provided data points.
 
         """
 
-        return(self._data_val)
+        return(dict(zip(self._data_idx, self._data_val)))
 
     @property
     def data_err(self):
         """
-        list of float: The upper and lower :math:`1\\sigma`-confidence levels
+        dict of float: The upper and lower :math:`1\\sigma`-confidence levels
         of provided data points.
 
         """
 
-        return(self._data_err)
+        return(dict(zip(self._data_idx, self._data_err)))
 
     @property
     def data_spc(self):
         """
-        list of str: The types of value space ({'lin', 'log', 'ln'}) of
+        dict of str: The types of value space ({'lin', 'log', 'ln'}) of
         provided data points.
 
         """
 
-        return(self._data_spc)
+        return(dict(zip(self._data_idx, self._data_spc)))
 
     @property
     def data_idx(self):
@@ -471,6 +468,42 @@ class ModelLink(object, metaclass=abc.ABCMeta):
 
         return(self._data_idx)
 
+    @property
+    def model_data(self):
+        """
+        dict of list: The dict of model data points as used by this
+        :obj:`~ModelLink` instance.
+
+        This dict can be used as the `model_data` argument when initializing
+        this :class:`~ModelLink` subclass.
+
+        """
+
+        # Initialize empty dict of model data
+        model_data = {}
+
+        # Combine data points together, only adding non-default values to dict
+        for idx, val, err, spc in zip(self._data_idx, self._data_val,
+                                      self._data_err, self._data_spc):
+            # Create data point
+            data_point = [val]
+
+            # Add data error, add only one value if error is centered
+            if(err[0] == err[1]):
+                data_point.append(err[0])
+            else:
+                data_point.extend(err)
+
+            # Add data space if it is not 'lin'
+            if(spc != 'lin'):
+                data_point.append(spc)
+
+            # Add data_point to model_data dict
+            model_data[idx] = data_point
+
+        # Return model_data
+        return(model_data)
+
     # %% GENERAL CLASS METHODS
     # This function returns non-default string representations of input args
     def get_str_repr(self):
@@ -533,7 +566,7 @@ class ModelLink(object, metaclass=abc.ABCMeta):
         logger.info("Converting sequence of model parameter names/indices.")
 
         # Remove all unwanted characters from the string and split it up
-        par_seq = convert_str_seq(par_seq)
+        par_seq = e13.split_seq(par_seq)
 
         # Check elements if they are ints or strings, and if they are valid
         for i, par_idx in enumerate(par_seq):
@@ -549,15 +582,15 @@ class ModelLink(object, metaclass=abc.ABCMeta):
             except Exception as error:
                 err_msg = ("Input argument %r[%i] is invalid! (%s)"
                            % (name, i, error))
-                raise_error(err_msg, InputError, logger)
+                e13.raise_error(err_msg, e13.InputError, logger)
 
         # If everything went without exceptions, check if list is not empty and
         # remove duplicates
-        if len(par_seq):
+        if par_seq:
             par_seq = list(sset(par_seq))
         else:
             err_msg = "Input argument %r is empty!" % (name)
-            raise_error(err_msg, ValueError, logger)
+            e13.raise_error(err_msg, ValueError, logger)
 
         # Log end
         logger.info("Finished converting sequence of model parameter "
@@ -566,6 +599,59 @@ class ModelLink(object, metaclass=abc.ABCMeta):
         # Return it
         return(par_seq)
 
+    # Returns the hypercube that encloses provided sam_set
+    def _get_sam_space(self, sam_set):
+        """
+        Returns the boundaries of the hypercube that encloses the parameter
+        space in which the provided `sam_set` is defined.
+
+        The main use for this function is to determine what part of model
+        parameter space was likely sampled from in order to obtain the provided
+        `sam_set`. Because of this, extra spacing is added to the boundaries to
+        reduce the effect of the used sampling method.
+
+        Parameters
+        ----------
+        sam_set : 1D or 2D array_like or dict
+            Parameter/sample set for which an enclosing hypercube is requested.
+
+        Returns
+        -------
+        sam_space : 2D :obj:`~numpy.ndarray` object
+            The requested hypercube boundaries.
+
+        """
+
+        # If sam_set is a dict, convert it to a NumPy array
+        if isinstance(sam_set, dict):
+            sam_set = np_array(sdict(sam_set).values()).T
+
+        # Make sure that sam_set is a 2D NumPy array
+        sam_set = np_array(sam_set, ndmin=2)
+
+        # Determine the maximum difference between consecutive samples
+        sam_diff = np.apply_along_axis(
+            lambda x: np.max(np.diff(np.sort(x))), axis=0, arr=sam_set)
+
+        # Determine the min/max values of all samples
+        sam_min = np.min(sam_set, axis=0)
+        sam_max = np.max(sam_set, axis=0)
+
+        # Add 3*sam_diff as extra spacing to sam_min and sam_max
+        # This reduces the effect of the used sampling method and randomness
+        sam_min -= 3*sam_diff
+        sam_max += 3*sam_diff
+
+        # Combine sam_min and sam_max to form sam_space
+        sam_space = np.stack([sam_min, sam_max], axis=1)
+
+        # Make sure that sam_space is within par_space
+        sam_space = np.apply_along_axis(
+            lambda x: np.clip(x, *self._par_rng.T), axis=0, arr=sam_space)
+
+        # Return sam_space
+        return(sam_space)
+
     # This function checks if a provided mod_set is valid
     def _check_mod_set(self, mod_set, name):
         """
@@ -585,7 +671,7 @@ class ModelLink(object, metaclass=abc.ABCMeta):
         mod_set : 1D or 2D :obj:`~numpy.ndarray` object
             The provided `mod_set` if the validation was successful. If
             `mod_set` was a dict, it will be converted to a
-            :obj:`~numpy.ndarray` object.
+            :obj:`~numpy.ndarray` object (sorted on :attr:`~data_idx`).
 
         """
 
@@ -593,9 +679,14 @@ class ModelLink(object, metaclass=abc.ABCMeta):
         logger = getCLogger('CHECK')
         logger.info("Validating provided set of model outputs %r." % (name))
 
-        # If mod_set is a dict, convert it to a NumPy array
+        # If mod_set is a dict, try to convert it to a NumPy array
         if isinstance(mod_set, dict):
-            mod_set = np_array([mod_set[idx] for idx in mod_set.keys()]).T
+            try:
+                mod_set = np_array([mod_set[idx] for idx in self._data_idx]).T
+            except KeyError as error:
+                err_msg = ("Input argument %r is missing data identifier '%r'!"
+                           % (name, error.args[0]))
+                e13.raise_error(err_msg, KeyError, logger)
 
         # Make sure that mod_set is a NumPy array
         mod_set = np_array(mod_set)
@@ -604,14 +695,14 @@ class ModelLink(object, metaclass=abc.ABCMeta):
         if not(mod_set.ndim == 1 or mod_set.ndim == 2):
             err_msg = ("Input argument %r is not one-dimensional or "
                        "two-dimensional!" % (name))
-            raise_error(err_msg, ShapeError, logger)
+            e13.raise_error(err_msg, e13.ShapeError, logger)
 
         # Raise error if mod_set does not have n_data data values
         if not(mod_set.shape[-1] == self._n_data):
             err_msg = ("Input argument %r has incorrect number of data values "
                        "(%i != %i)!"
                        % (name, mod_set.shape[-1], self._n_data))
-            raise_error(err_msg, ShapeError, logger)
+            e13.raise_error(err_msg, e13.ShapeError, logger)
 
         # Check if mod_set solely consists out of floats
         mod_set = check_vals(mod_set, name, 'float')
@@ -661,14 +752,14 @@ class ModelLink(object, metaclass=abc.ABCMeta):
         if not(sam_set.ndim == 1 or sam_set.ndim == 2):
             err_msg = ("Input argument %r is not one-dimensional or "
                        "two-dimensional!" % (name))
-            raise_error(err_msg, ShapeError, logger)
+            e13.raise_error(err_msg, e13.ShapeError, logger)
 
         # Raise error if sam_set does not have n_par parameter values
         if not(sam_set.shape[-1] == self._n_par):
             err_msg = ("Input argument %r has incorrect number of parameters "
                        "(%i != %i)!"
                        % (name, sam_set.shape[-1], self._n_par))
-            raise_error(err_msg, ShapeError, logger)
+            e13.raise_error(err_msg, e13.ShapeError, logger)
 
         # Check if sam_set solely consists out of floats
         sam_set = check_vals(sam_set, name, 'float')
@@ -687,13 +778,47 @@ class ModelLink(object, metaclass=abc.ABCMeta):
         else:
             err_msg = ("Input argument '%s%s' is outside parameter space!"
                        % (name, index if sam_set.ndim != 1 else ''))
-            raise_error(err_msg, ValueError, logger)
+            e13.raise_error(err_msg, ValueError, logger)
 
         # Log again and return sam_set
         logger.info("Finished validating provided set of model parameter "
                     "samples %r." % (name))
         return(sam_set)
 
+    # This function converts a given sam_set to a sam_dict
+    def _get_sam_dict(self, sam_set):
+        """
+        Converts a provided set of model parameter samples `sam_set` to a dict
+        for use in this :obj:`~ModelLink` instance.
+
+        This dict can be used as the `par_set` argument in the
+        :meth:`~call_model` and :meth:`~get_md_var` methods.
+
+        Parameters
+        ----------
+        sam_set : 1D or 2D array_like
+            Parameter/sample set to convert for this :obj:`~ModelLink`
+            instance.
+
+        Returns
+        -------
+        sam_dict : dict of list
+            Dict of parameter samples.
+
+        """
+
+        # Make sure that sam_set is a NumPy array
+        sam_set = np_array(sam_set)
+
+        # Check how many dimensions sam_set has and act accordingly
+        if(sam_set.ndim == 1):
+            sam_dict = sdict(zip(self._par_name, sam_set))
+        else:
+            sam_dict = sdict(zip(self._par_name, sam_set.T))
+
+        # Return sam_dict
+        return(sam_dict)
+
     # This function checks if a provided md_var is valid
     def _check_md_var(self, md_var, name):
         """
@@ -734,14 +859,14 @@ class ModelLink(object, metaclass=abc.ABCMeta):
         if not(md_var.ndim == 1 or md_var.ndim == 2):
             err_msg = ("Input argument %r is not one-dimensional or "
                        "two-dimensional!" % (name))
-            raise_error(err_msg, ShapeError, logger)
+            e13.raise_error(err_msg, e13.ShapeError, logger)
 
         # Check if md_var contains n_data values
         if not(md_var.shape[0] == self._n_data):
             err_msg = ("Received array of model discrepancy variances %r has "
                        "incorrect number of data points (%i != %i)!"
                        % (name, md_var.shape[0], self._n_data))
-            raise ShapeError(err_msg)
+            raise e13.ShapeError(err_msg)
 
         # Check if single or dual values were given
         if(md_var.ndim == 1):
@@ -752,7 +877,7 @@ class ModelLink(object, metaclass=abc.ABCMeta):
             err_msg = ("Received array of model discrepancy variances %r has "
                        "incorrect number of values (%i != 2)!"
                        % (name, md_var.shape[1]))
-            raise ShapeError(err_msg)
+            raise e13.ShapeError(err_msg)
 
         # Check if all values are non-negative floats
         md_var = check_vals(md_var, 'md_var', 'nneg', 'float')
@@ -802,30 +927,13 @@ class ModelLink(object, metaclass=abc.ABCMeta):
         # If suffix is a string, determine the path
         elif isinstance(suffix, str):
             # If the string is empty, find the last created backup file
-            if(suffix == ''):
-                # Make list of all files in current directory
-                filenames = next(os.walk('.'))[2]
+            if not suffix:
+                # Make list of all valid backup files in current directory
+                files = glob("%s/%s*" % (path.abspath('.'), prefix))
 
-                # Make empty list of all backup files
-                backup_files = []
-
-                # Loop over all filenames
-                for filename in filenames:
-                    # If the filename has the correct prefix
-                    if filename.startswith(prefix):
-                        # Obtain full path to the file
-                        filepath = path.abspath(path.join('.', filename))
-
-                        # Obtain creation time and append to backup_files
-                        ctime = path.getctime(filepath)
-                        backup_files.append([filepath, ctime])
-
-                # Sort backup_files list on creation time
-                backup_files.sort(key=lambda x: x[1], reverse=True)
-
-                # If backup_files is not empty, return last one created
-                if len(backup_files):
-                    return(backup_files[0][0])
+                # If files is not empty, return last one created
+                if files:
+                    return(max(files, key=path.getctime))
                 # Else, raise error
                 else:
                     err_msg = ("No backup files can be found in the current "
@@ -836,7 +944,7 @@ class ModelLink(object, metaclass=abc.ABCMeta):
             else:
                 # Obtain full filepath
                 filepath = path.abspath(path.join(
-                        '.', ''.join([prefix, suffix, ').hdf5'])))
+                    '.', ''.join([prefix, suffix, ').hdf5'])))
 
                 # If filepath exists, return it
                 if path.exists(filepath):
@@ -868,9 +976,9 @@ class ModelLink(object, metaclass=abc.ABCMeta):
 
         Parameters
         ----------
-        args : tuple
+        args : positional arguments
             All positional arguments that must be stored in the backup file.
-        kwargs : dict
+        kwargs : keyword arguments
             All keyword arguments that must be stored in the backup file.
 
         Notes
@@ -902,7 +1010,7 @@ class ModelLink(object, metaclass=abc.ABCMeta):
             return
 
         # Obtain the call_model frame
-        caller_frame = get_outer_frame(self.call_model)
+        caller_frame = e13.get_outer_frame(self.call_model)
 
         # If caller_frame is None, the call_model frame was not found
         if caller_frame is None:
@@ -954,6 +1062,7 @@ class ModelLink(object, metaclass=abc.ABCMeta):
 
     # This function reads in a backup made by _make_backup
     # TODO: Allow for absolute path to backup file to be given?
+    # TODO: Convert to static method to read backups without subclass object?
     def _read_backup(self, emul_i, *, suffix=None):
         """
         Reads in a backup HDF5-file created by the :meth:`~_make_backup`
@@ -969,7 +1078,8 @@ class ModelLink(object, metaclass=abc.ABCMeta):
         --------
         suffix : str or None. Default: None
             The suffix of the backup file (everything between parentheses) that
-            needs to be read. If *None*, the last created backup will be read.
+            needs to be read. If *None* or empty, the last created backup will
+            be read.
 
         Returns
         -------
@@ -1074,7 +1184,8 @@ class ModelLink(object, metaclass=abc.ABCMeta):
         # Save number of model parameters
         n_par = len(model_parameters.keys())
         if(n_par == 1):
-            raise InputError("Number of model parameters must be at least 2!")
+            raise e13.InputError("Number of model parameters must be at least "
+                                 "2!")
         else:
             self._n_par = check_vals(n_par, 'n_par', 'pos')
 
@@ -1166,7 +1277,7 @@ class ModelLink(object, metaclass=abc.ABCMeta):
 
     # %% ABSTRACT USER METHODS
     @abc.abstractmethod
-    @docstring_substitute(emul_i=std_emul_i_doc)
+    @e13.docstring_substitute(emul_i=std_emul_i_doc)
     def call_model(self, emul_i, par_set, data_idx):
         """
         Calls the model wrapped in this :class:`~ModelLink` subclass at
@@ -1200,6 +1311,12 @@ class ModelLink(object, metaclass=abc.ABCMeta):
             If dict, it has the identifiers in `data_idx` as its keys with
             either scalars or 1D array_likes as its values.
 
+        Note
+        ----
+        If this model is multi-called, then the parameter sets in the provided
+        `par_set` dict will be sorted in order of parameter name (e.g., sort on
+        first parameter first, then on second parameter, etc.).
+
         """
 
         # Raise NotImplementedError if only super() was called
@@ -1207,7 +1324,7 @@ class ModelLink(object, metaclass=abc.ABCMeta):
                                   "ModelLink subclass!")
 
     @abc.abstractmethod
-    @docstring_substitute(emul_i=std_emul_i_doc)
+    @e13.docstring_substitute(emul_i=std_emul_i_doc)
     def get_md_var(self, emul_i, par_set, data_idx):
         """
         Calculates the linear model discrepancy variance at a given emulator
