@@ -15,11 +15,11 @@ from inspect import isfunction
 import warnings
 
 # Package imports
-from e13tools import InputError
-from e13tools.sampling import lhd
-from e13tools.utils import docstring_substitute
+import e13tools as e13
 import numpy as np
+from numpy.random import multivariate_normal
 from sortedcontainers import SortedDict as sdict
+from tqdm import tqdm
 
 # PRISM imports
 from prism._docstrings import user_emul_i_doc
@@ -27,12 +27,12 @@ from prism._internal import RequestError, check_vals, np_array
 from prism._pipeline import Pipeline
 
 # All declaration
-__all__ = ['get_hybrid_lnpost_fn', 'get_walkers', 'get_lnpost_fn']
+__all__ = ['get_hybrid_lnpost_fn', 'get_walkers']
 
 
 # %% FUNCTION DEFINITIONS
 # This function returns a hybrid version of the lnpost function
-@docstring_substitute(emul_i=user_emul_i_doc)
+@e13.docstring_substitute(emul_i=user_emul_i_doc)
 def get_hybrid_lnpost_fn(lnpost_fn, pipeline_obj, *, emul_i=None,
                          unit_space=False, impl_prior=True, par_dict=False):
     """
@@ -102,8 +102,8 @@ def get_hybrid_lnpost_fn(lnpost_fn, pipeline_obj, *, emul_i=None,
 
     # Check if lnpost_fn is a function
     if not isfunction(lnpost_fn):
-        raise InputError("Input argument 'lnpost_fn' is not a callable "
-                         "function definition!")
+        raise e13.InputError("Input argument 'lnpost_fn' is not a callable "
+                             "function definition!")
 
     # Make abbreviation for pipeline_obj
     pipe = pipeline_obj
@@ -115,11 +115,11 @@ def get_hybrid_lnpost_fn(lnpost_fn, pipeline_obj, *, emul_i=None,
 
     # Check if the provided pipeline_obj uses a default emulator
     if(pipe._emulator._emul_type != 'default'):
-        raise InputError("Input argument 'pipeline_obj' does not use a default"
-                         " emulator!")
+        raise e13.InputError("Input argument 'pipeline_obj' does not use a "
+                             "default emulator!")
 
     # Get emulator iteration
-    emul_i = pipe._emulator._get_emul_i(emul_i, True)
+    emul_i = pipe._emulator._get_emul_i(emul_i)
 
     # Check if unit_space is a bool
     unit_space = check_vals(unit_space, 'unit_space', 'bool')
@@ -149,9 +149,9 @@ def get_hybrid_lnpost_fn(lnpost_fn, pipeline_obj, *, emul_i=None,
             Sample to calculate the posterior probability for. This sample is
             first analyzed in `pipeline_obj` and only given to `lnpost_fn` if
             it is plausible. If `par_dict` is *True*, this is a dict.
-        args : tuple
+        args : positional arguments
             Positional arguments that need to be passed to `lnpost_fn`.
-        kwargs : dict
+        kwargs : keyword arguments
             Keyword arguments that need to be passed to `lnpost_fn`.
 
         Returns
@@ -208,16 +208,17 @@ def get_hybrid_lnpost_fn(lnpost_fn, pipeline_obj, *, emul_i=None,
 
 
 # This function returns a set of valid MCMC walkers
-@docstring_substitute(emul_i=user_emul_i_doc)
+@e13.docstring_substitute(emul_i=user_emul_i_doc)
 def get_walkers(pipeline_obj, *, emul_i=None, init_walkers=None,
-                unit_space=False, lnpost_fn=None, **kwargs):
+                req_n_walkers=None, unit_space=False, lnpost_fn=None,
+                **kwargs):
     """
     Analyzes proposed `init_walkers` and returns plausible `p0_walkers`.
 
     Analyzes sample set `init_walkers` in the provided `pipeline_obj` at
     iteration `emul_i` and returns all samples that are plausible to be used as
-    MCMC walkers. The provided samples and returned walkers should be/are given
-    in unit space if `unit_space` is *True*.
+    starting positions for MCMC walkers. The provided samples and returned
+    walkers should be/are given in unit space if `unit_space` is *True*.
 
     If `init_walkers` is *None*, returns :attr:`~prism.Pipeline.impl_sam`
     instead if it is available.
@@ -228,7 +229,8 @@ def get_walkers(pipeline_obj, *, emul_i=None, init_walkers=None,
     ----------
     pipeline_obj : :obj:`~prism.Pipeline` object
         The instance of the :class:`~prism.Pipeline` class that needs to be
-        used for determining the plausibility of the proposed walkers.
+        used for determining the plausibility of the proposed starting
+        positions.
 
     Optional
     --------
@@ -240,6 +242,12 @@ def get_walkers(pipeline_obj, *, emul_i=None, init_walkers=None,
         samples.
         If *None*, return :attr:`~prism.Pipeline.impl_sam` corresponding to
         iteration `emul_i` instead.
+    req_n_walkers : int or None. Default: None
+        The minimum required number of plausible starting positions that should
+        be returned. If *None*, all plausible starting positions in
+        `init_walkers` are returned instead.
+
+        .. versionadded:: 1.2.0
     unit_space : bool. Default: False
         Bool determining whether or not the provided samples and returned
         walkers are given in unit space.
@@ -252,7 +260,8 @@ def get_walkers(pipeline_obj, *, emul_i=None, init_walkers=None,
     Returns
     -------
     n_walkers : int
-        Number of returned MCMC walkers.
+        Number of returned MCMC walkers. Note that this number can be higher
+        than `req_n_walkers` if not *None*.
     p0_walkers : 2D :obj:`~numpy.ndarray` object or dict
         Array containing plausible starting positions of valid MCMC walkers.
         If `init_walkers` was provided as a dict, `p0_walkers` will be a dict.
@@ -276,6 +285,13 @@ def get_walkers(pipeline_obj, *, emul_i=None, init_walkers=None,
     If `init_walkers` is *None* and emulator iteration `emul_i` has not been
     analyzed yet, a :class:`~prism._internal.RequestError` will be raised.
 
+    If `req_n_walkers` is not *None*, a custom Metropolis-Hastings sampling
+    algorithm is used to generate the required number of starting positions.
+    All plausible samples in `init_walkers` are used as the start of every MCMC
+    chain. Note that if the number of plausible samples in `init_walkers` is
+    small, it is possible that the returned `p0_walkers` are not spread out
+    properly over parameter space.
+
     """
 
     # Make abbreviation for pipeline_obj
@@ -288,11 +304,16 @@ def get_walkers(pipeline_obj, *, emul_i=None, init_walkers=None,
 
     # Check if the provided pipeline_obj uses a default emulator
     if(pipe._emulator._emul_type != 'default'):
-        raise InputError("Input argument 'pipeline_obj' does not use a default"
-                         " emulator!")
+        raise e13.InputError("Input argument 'pipeline_obj' does not use a "
+                             "default emulator!")
 
     # Get emulator iteration
-    emul_i = pipe._emulator._get_emul_i(emul_i, True)
+    emul_i = pipe._emulator._get_emul_i(emul_i)
+
+    # If req_n_walkers is not None, check if it is an integer
+    if req_n_walkers is not None:
+        req_n_walkers = check_vals(req_n_walkers, 'req_n_walkers', 'int',
+                                   'pos')
 
     # Check if unit_space is a bool
     unit_space = check_vals(unit_space, 'unit_space', 'bool')
@@ -306,8 +327,8 @@ def get_walkers(pipeline_obj, *, emul_i=None, init_walkers=None,
             hybrid_lnpost =\
                 get_hybrid_lnpost_fn(lnpost_fn, pipe, emul_i=emul_i,
                                      unit_space=unit_space, **kwargs)
-        except InputError:
-            raise InputError("Input argument 'lnpost_fn' is invalid!")
+        except e13.InputError:
+            raise e13.InputError("Input argument 'lnpost_fn' is invalid!")
 
     # If init_walkers is None, use impl_sam of emul_i
     if init_walkers is None:
@@ -339,10 +360,17 @@ def get_walkers(pipeline_obj, *, emul_i=None, init_walkers=None,
                 # Check if provided integer is positive
                 n_sam = check_vals(init_walkers, 'init_walkers', 'pos')
 
+                # Obtain the par_space to sample in
+                par_space = pipe._get_impl_space(emul_i)
+
+                # If par_space is None, use the corresponding emul_space
+                if par_space is None:
+                    par_space = pipe._emulator._emul_space[emul_i]
+
                 # Create LHD of provided size
-                init_walkers = lhd(n_sam, pipe._modellink._n_par,
-                                   pipe._modellink._par_rng, 'center',
-                                   pipe._criterion, 100)
+                init_walkers = e13.lhd(n_sam, pipe._modellink._n_par,
+                                       par_space, 'center', pipe._criterion,
+                                       100)
 
             # If init_walkers is not an int, it must be array_like or dict
             else:
@@ -374,13 +402,17 @@ def get_walkers(pipeline_obj, *, emul_i=None, init_walkers=None,
         # Analyze init_walkers and save them as p0_walkers
         p0_walkers = pipe._evaluate_sam_set(emul_i, init_walkers, 'analyze')
 
-    # Calculate n_walkers
-    n_walkers = len(p0_walkers)
+    # Check if init_walkers is not empty and raise error if it is
+    if not p0_walkers.shape[0]:
+        raise e13.InputError("Input argument 'init_walkers' contains no "
+                             "plausible samples!")
 
-    # Check if p0_walkers is not empty
-    if not n_walkers:
-        raise InputError("Input argument 'init_walkers' contains no plausible "
-                         "samples!")
+    # If req_n_walkers is not None, use MH MCMC to find all required walkers
+    if req_n_walkers is not None:
+        n_walkers, p0_walkers = _do_mh_walkers(pipe, p0_walkers, req_n_walkers)
+    else:
+        p0_walkers = np.unique(p0_walkers, axis=0)
+        n_walkers = p0_walkers.shape[0]
 
     # Check if p0_walkers needs to be converted
     if unit_space:
@@ -388,7 +420,7 @@ def get_walkers(pipeline_obj, *, emul_i=None, init_walkers=None,
 
     # Check if p0_walkers needs to be returned as a dict
     if walker_dict:
-        p0_walkers = sdict(zip(pipe._modellink._par_name, p0_walkers.T))
+        p0_walkers = pipe._modellink._get_sam_dict(p0_walkers)
 
     # Check if hybrid_lnpost was requested and return it as well if so
     if lnpost_fn is not None:
@@ -397,85 +429,115 @@ def get_walkers(pipeline_obj, *, emul_i=None, init_walkers=None,
         return(n_walkers, p0_walkers)
 
 
-# %% DEPRECATED FUNCTION DEFINITIONS
-# This function returns a specialized version of the lnpost function
-@docstring_substitute(emul_i=user_emul_i_doc)
-def get_lnpost_fn(ext_lnpost, pipeline_obj, *, emul_i=None, unit_space=True,
-                  hybrid=True, par_dict=False):  # pragma: no cover
+# %% HIDDEN FUNCTION DEFINITIONS
+# This function uses MH sampling to find req_n_walkers initial positions
+def _do_mh_walkers(pipeline_obj, p0_walkers, req_n_walkers):
     """
-    .. deprecated:: 1.1.3
-
-    Returns a function definition ``get_lnpost(par_set, *args, **kwargs)``.
-
-    This `get_lnpost` function can be used to calculate the natural logarithm
-    of the posterior probability, which analyzes a given `par_set` first in the
-    provided `pipeline_obj` at iteration `emul_i` and passes it to the
-    `ext_lnpost` function if it is plausible.
+    Generates `req_n_walkers` unique starting positions for MCMC walkers using
+    Metropolis-Hastings sampling and the provided `pipeline_obj` and
+    `p0_walkers`.
 
     This function needs to be called by all MPI ranks.
 
     Parameters
     ----------
-    ext_lnpost : function
-        Function definition that needs to be called if the provided `par_set`
-        is plausible in iteration `emul_i` of `pipeline_obj`. The used call
-        signature is ``ext_lnpost(par_set, *args, **kwargs)``. All MPI ranks
-        will call this function unless called within the
-        :attr:`~prism.Pipeline.worker_mode` context manager.
     pipeline_obj : :obj:`~prism.Pipeline` object
-        The instance of the :class:`~prism.Pipeline` class that needs
-        to be used for determining the validity of the proposed sampling step.
-
-    Optional
-    --------
-    %(emul_i)s
-    unit_space : bool. Default: True
-        Bool determining whether or not `par_set` will be given in unit space.
-    hybrid : bool. Default: True
-        Bool determining whether or not the `get_lnpost` function should
-        use the implausibility values of a given `par_set` as an additional
-        prior.
-    par_dict : bool. Default: False
-        Bool determining whether or not `par_set` will be an array_like
-        (*False*) or a dict (*True*).
+        The instance of the :class:`~prism.Pipeline` class that needs to be
+        used for determining the validity of a proposed sampling step.
+    p0_walkers : 2D :obj:`~numpy.ndarray` object
+        Sample set of starting positions for the MCMC chains.
+    req_n_walkers : int
+        The minimum required number of unique MCMC walker positions that must
+        be returned.
 
     Returns
     -------
-    get_lnpost : function
-        Definition of the function ``get_lnpost(par_set, *args, **kwargs)``.
-
-    See also
-    --------
-    :func:`~get_walkers`
-        Analyzes proposed `init_walkers` and returns valid `p0_walkers`.
-
-    :attr:`~prism.Pipeline.worker_mode`
-        Special context manager within which all code is executed in worker
-        mode.
+    n_walkers : int
+        Number of unique MCMC walker positions that are returned.
+    walkers : 2D :obj:`~numpy.ndarray` object
+        Array containing all unique MCMC walker positions.
 
     Note
     ----
-    The input arguments `unit_space` and `par_dict` state in what form
-    `par_set` will be provided to the ``get_lnpost()`` function, such that it
-    can be properly converted to the format used in :class:`~prism.Pipeline`.
-    The `par_set` that is passed to the ``ext_lnpost()`` function is unchanged.
-
-    Warning
-    -------
-    Calling this function factory will disable all regular logging in
-    `pipeline_obj` (:attr:`~prism.Pipeline.do_logging` set to
-    *False*), in order to avoid having the same message being logged every time
-    `get_lnpost` is called.
+    Executing this function will temporarily disable all regular logging in the
+    provided :obj:`~prism.Pipeline` object. If logging was enabled before this
+    function was executed, it will be enabled again afterward.
 
     """
 
-    # Raise a FutureWarning
-    warn_msg = ("This function factory was remade into 'get_hybrid_lnpost_fn' "
-                "in v1.1.3. This compatibility definition will be removed "
-                "entirely in v1.2.0.")
-    warnings.warn(warn_msg, FutureWarning, stacklevel=2)
+    # Make abbreviation for pipeline_obj
+    pipe = pipeline_obj
 
-    # Call new get_hybrid_lnpost_fn() function factory
-    return(get_hybrid_lnpost_fn(ext_lnpost, pipeline_obj, emul_i=emul_i,
-                                unit_space=unit_space, impl_prior=hybrid,
-                                par_dict=par_dict))
+    # Define function to check if proposed sam_set should be accepted
+    def advance_chain(sam_set):
+        # Make sure that sam_set is 2D
+        sam_set = np_array(sam_set, ndmin=2)
+
+        # Check if sam_set is within parameter space and reject if not
+        par_rng = pipe._modellink._par_rng
+        accept = ((par_rng[:, 0] <= sam_set)*(sam_set <= par_rng[:, 1])).all(1)
+
+        # Evaluate all non-rejected samples and accept if plausible
+        emul_i = pipe._emulator._emul_i
+        accept[accept] = pipe._make_call('_evaluate_sam_set', emul_i,
+                                         sam_set[accept], 'project')[0]
+
+        # Return which samples should be accepted or rejected
+        return(accept)
+
+    # Initialize array of final walkers
+    n_walkers = p0_walkers.shape[0]
+    walkers = np.empty([req_n_walkers+n_walkers-1, pipe._modellink._n_par])
+    walkers[:n_walkers] = p0_walkers
+
+    # Check if logging is currently turned on
+    was_logging = bool(pipe.do_logging)
+
+    # Make sure that logging is turned off
+    pipe.do_logging = False
+
+    # Use worker mode
+    with pipe.worker_mode:
+        if pipe._is_controller:
+            # Initialize progress bar
+            pbar = tqdm(desc="Finding walkers", total=req_n_walkers,
+                        initial=n_walkers, disable=not was_logging,
+                        bar_format=("{l_bar}{bar}| {n_fmt}/{total_fmt} "
+                                    "[Time: {elapsed}]"))
+
+            # Keep searching parameter space until req_n_walkers are found
+            while(n_walkers < req_n_walkers):
+                # Calculate the covariance matrix of all walkers
+                cov = np.cov(walkers[:n_walkers].T)
+
+                # Create set of proposed walkers
+                new_walkers = np.apply_along_axis(multivariate_normal, 1,
+                                                  p0_walkers, cov)
+
+                # Check which proposed walkers should be accepted
+                accept = advance_chain(new_walkers)
+                acc_walkers = new_walkers[accept]
+                n_accepted = sum(accept)
+
+                # Replace current walkers with accepted walkers
+                p0_walkers[accept] = acc_walkers
+
+                # Update final walkers array
+                walkers[n_walkers:n_walkers+n_accepted] = acc_walkers
+                n_walkers += n_accepted
+
+                # Update progress bar
+                pbar.update(min(req_n_walkers, n_walkers)-pbar.n)
+
+            # Close progress bar
+            pbar.close()
+
+    # Turn logging back on if it used to be on
+    pipe.do_logging = was_logging
+
+    # Broadcast walkers to all workers
+    walkers = pipe._comm.bcast(np.unique(walkers[:req_n_walkers], axis=0), 0)
+    n_walkers = walkers.shape[0]
+
+    # Return n_walkers, walkers
+    return(n_walkers, walkers)
